/*****************************************************************************
* Product: DPP example
* Last Updated for Version: 4.0.00
* Date of the Last Update:  Apr 07, 2008
*
*                    Q u a n t u m     L e a P s
*                    ---------------------------
*                    innovating embedded systems
*
* Copyright (C) 2002-2008 Quantum Leaps, LLC. All rights reserved.
*
* This software may be distributed and modified under the terms of the GNU
* General Public License version 2 (GPL) as published by the Free Software
* Foundation and appearing in the file GPL.TXT included in the packaging of
* this file. Please note that GPL Section 2[b] requires that all works based
* on this software must also be made publicly available under the terms of
* the GPL ("Copyleft").
*
* Alternatively, this software may be distributed and modified under the
* terms of Quantum Leaps commercial licenses, which expressly supersede
* the GPL and are specifically designed for licensees interested in
* retaining the proprietary status of their code.
*
* Contact information:
* Quantum Leaps Web site:  http://www.quantum-leaps.com
* e-mail:                  info@quantum-leaps.com
*****************************************************************************/
#ifdef HAVE_CONFIG_H
    #include "config.h"
#endif

#ifdef HAVE_AVAHI
    #include <libconfig.h>
    #include <sys/stat.h>
#endif

#include <errno.h>
#include <limits.h>
#include <stdlib.h>
#include <syslog.h>
#include <unistd.h>

#include "qp_port.h"
#include "bsp.h"
#include "dsm.h"
#include "receiver.h"
#include "datamodel.h"
#include "dsmapi.h"
#include "selectB.h"
#include "selectA.h"
#include "mmi.h"
#include "util.h"
#include "uthash.h"
#include "vdc-api-proto.h"
#include "role.h"

extern int optind;

static QSubscrList l_subscrSto[MAX_PUBLISH_SIG];
/* TODO: figure out minimum queue sizes */
static QEvent const * l_ds485QueueSto[128];      /* event queue storage */
static QEvent const * l_datamodelQueueSto[128];  /* event queue storage */
static QEvent const * l_dsmApiQueueSto[128];     /* event queue storage */
static QEvent const * l_selectBQueueSto[128];    /* event queue storage */
static QEvent const * l_selectAQueueSto[128];    /* event queue storage */
static QEvent const * l_mmiQueueSto[128];        /* event queue storage */

typedef union {
    FrontPanelEvt frontPanel;
} SmallPoolUnion;

typedef union {
    DS485ShortMessageEvt ds485;
    DatamodelReqSetNameEvt setName;
} MediumPoolUnion;

typedef union {
    DS485MessageEvt ds485;
    DS485NetlibPacketMessageEvt ds485nPacket;
    DatamodelResDeviceEntryEvt resDeviceEntry;
} LargePoolUnion;

static SmallPoolUnion l_smallPoolSto[128];
static MediumPoolUnion l_mediumPoolSto[128];
static LargePoolUnion l_largePoolSto[128];

#define POWER_DOWN_MAGIC 0xFACE


/*
 * Command line options
 */
static bool listenGlobalDS485 = FALSE;
static char* listenPortDS485 = "8441";

static char* databasePath = "./vdsm.db";
static dsuid_t vDSMDsuid = { { 0x35, 0x04, 0x17, 0x5f, 0xe0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x10, 0x00, 0x00, 0x00, 0x01, 0x00 } };

static int debugLevel = 5;


static void print_usage(void)
{
    fprintf(stderr, "Usage: vdsm [OPTION] [<uri> ...]\n\n");
    fprintf(stderr, "       <uri> is a vDC listening socket (default: localhost:8440)\n\n");
    fprintf(stderr, "       OPTION is a selection out of the following options:\n");
    fprintf(stderr, "       -h  --  print this help\n");
    fprintf(stderr, "       -l  --  debug-level: max debug level still logged: 0-7 (default: 5)\n");
    fprintf(stderr, "       -b  --  dS485 interface listening port (default: 8441)\n");
    fprintf(stderr, "       -B  --  enable external access: accept ds485 socket connection from other machines\n");
    fprintf(stderr, "       -u  --  collectable: can be automatically connected to from an upstream dS485p\n");
    fprintf(stderr, "       -n  --  noauto: Request the ds485p not to connect unless explicitly requested by the user\n");
    fprintf(stderr, "       -d  --  collecting: can connect to downstream vDCs automatically\n");
    fprintf(stderr, "       -M  --  master: announce as master to other vDSMs\n");
    fprintf(stderr, "       -a  --  auxiliary: disable automatically in the presence of a separate master vDSM\n");
    fprintf(stderr, "       -m  --  change dSUID of vdSM (default: 3504175fe0000000000000100000000100\n");
    fprintf(stderr, "       -s  --  select path to vdsm database (default: ./vdsm.db)\n");
    fprintf(stderr, "       -c  --  config file path (default: ./vdsm.cfg)\n");
    fprintf(stderr, "       -V  --  print vdSM version information\n");
    fprintf(stderr, "\n");
}

#ifdef HAVE_AVAHI
static void prepare_whitelist(const char *cfg_file,
                              vdc_whitelist_hash_t **whitelist)
{
    struct stat statbuf;
    config_t config;
    const config_setting_t *devices;

    *whitelist = NULL;

    if (!cfg_file) {
        return;
    }

    if (stat(cfg_file, &statbuf) != 0) {
        vdsm_report(LOG_WARNING, "could not open configuration file %s\n", cfg_file);
        return;
    }

    config_init(&config);
    if (!config_read_file(&config, cfg_file)) {
        vdsm_report(LOG_ERR, "error in %s: %d %s\n", cfg_file,
               config_error_line(&config), config_error_text(&config));
        config_destroy(&config);
        return;
    }

    devices = config_lookup(&config, "autoconnect-whitelist");
    if (!devices) {
        vdsm_report(LOG_INFO, "config file does not specify autoconnect-whitelist "
                         "setting, all devices allowed.\n");
        config_destroy(&config);
        return;
    }

    if (!config_setting_is_array(devices)) {
        vdsm_report(LOG_ERR, "error in configuration, autoconnect-whitelist "
                        "must be an array\n");
        config_destroy(&config);
        return;
    }

    int i;
    for (i = 0; i < config_setting_length(devices); i++) {
        vdc_whitelist_hash_t *vdc_whitelist;
        // prevent double entries
        const char *dsuid = config_setting_get_string_elem(devices, i);
        if (!dsuid) {
            continue;
        }

        HASH_FIND_STR(*whitelist, dsuid, vdc_whitelist);
        if (vdc_whitelist == NULL) {
            vdc_whitelist = malloc(sizeof(vdc_whitelist_hash_t));
            if (!vdc_whitelist) {
                vdsm_report(LOG_ERR, "could not allocate memory for autoconnect "
                                "whitelist item\n");
                continue;
            }

            strncpy(vdc_whitelist->dsuid, dsuid, DSUID_STR_LEN-1);
            vdc_whitelist->dsuid[DSUID_STR_LEN-1] = '\0';
            HASH_ADD_STR(*whitelist, dsuid, vdc_whitelist);
            vdsm_report(LOG_DEBUG, "added device %s to whitelist\n",
                  vdc_whitelist->dsuid);
        }
    }

    config_destroy(&config);
}
#endif

int main(int argc, char *argv[])
{
    uint8_t prio;
    bool longPowerDown = TRUE;
    int i, c;
#ifdef HAVE_AVAHI
    const char *cfg = "vdsm.cfg";
#endif
    VDSMRole role;
    vdsm_role_clear(&role);

    /* command line parsing */
    if (argc < 1)
    {
        print_usage();
        exit(-1);
    }

    /* Look for options */
    while ((c = getopt(argc, argv, "hl:uBb:ndMas:m:c:V")) != -1)
    {
        switch (c)
        {
            case 'h':
                print_usage();
                exit(0);
            case 'l':
                debugLevel = atoi(optarg);
                break;
            case 'u':
                role.collectable = true;
                break;
            case 'B':
                listenGlobalDS485 = 1;
                break;
            case 'b':
                listenPortDS485 = strdup(optarg);
                break;
            case 'd':
                role.collecting = true;
                break;
            case 'M':
                role.master = true;
                break;
            case 'a':
                role.auxiliary = true;
                break;
            case 'n':
                role.noauto = true;
                break;
            case 's':
                databasePath = strdup(optarg);
                break;
#ifdef HAVE_AVAHI
            case 'c':
                cfg = optarg;
                break;
#endif
            case 'm':
            {
                const char *dsuid_str = optarg;
                if (strlen(dsuid_str) != (DSUID_STR_LEN - 1))
                {
                    vdsm_report(LOG_ERR, "vDSM dSUID in the wrong format\n");
                    exit(-1);
                }
                for (i = 0; i < DSUID_SIZE; i++)
                {
                    sscanf(&dsuid_str[i * 2], "%2hhx", &vDSMDsuid.id[i]);
                }
                break;
            }

            case 'V':
                printf("Version: v%s\n", VERSION);
                exit(0);
            default:
                exit(-1);
        }
    }
    if (!vdsm_role_validate(&role)) {
        exit(-1);
    }

    /* only set the default persistent connection if autodiscovery is
       disabled */
#ifndef HAVE_AVAHI
    /* parse connection arguments */
    if (optind >= argc) {
        /* no connection specified -> use default uri */
        vdc_api_add_connection("localhost:8440");
    } else {
#endif
        /* parse connection arguments */
        while (optind < argc) {
            const char *vDcConnection;
            vDcConnection = argv[optind];
            vdc_api_add_connection(vDcConnection);
            ++optind;
        }
#ifndef HAVE_AVAHI
    }
#endif

    vdsm_init_report(debugLevel);

    /* initialize board support package */
    BSP_init();

    /* initialize the framework and the underlying RT kernel */
    QF_init();

    /* initialize publish-subscribe */
    QF_psInit(l_subscrSto, Q_DIM(l_subscrSto));

    /* instantiate active objects */
    DS485_ctor(listenPortDS485, listenGlobalDS485);
    Datamodel_ctor(vDSMDsuid, databasePath);
    DsmApi_ctor(longPowerDown);
    SelectA_ctor();
    SelectB_ctor();
    MMI_ctor();

    /* initialize the event pools */
    QF_poolInit(l_smallPoolSto, sizeof(l_smallPoolSto), sizeof(l_smallPoolSto[0]));
    QF_poolInit(l_mediumPoolSto, sizeof(l_mediumPoolSto), sizeof(l_mediumPoolSto[0]));
    QF_poolInit(l_largePoolSto, sizeof(l_largePoolSto), sizeof(l_largePoolSto[0]));

    /* top state function dictionary entry */
    QS_FUN_DICTIONARY(&QHsm_top);
    /* published signals dictionary entries */
    QS_SIG_DICTIONARY(START_SIG, 0);
    QS_SIG_DICTIONARY(CONFIG_INIT_SIG, 0);
    QS_SIG_DICTIONARY(DS485_ADDRESS_SIG, 0);
    QS_SIG_DICTIONARY(SHORT_UPSTREAM_SIG, 0);
    QS_SIG_DICTIONARY(LONG_UPSTREAM_SIG, 0);
    QS_SIG_DICTIONARY(DOWNSTREAM_DONE_SIG, 0);
    QS_SIG_DICTIONARY(SYSTEM_TERM_SIG, 0);

    /* start the active objects */
    prio = 0;
    QActive_start(AO_MMI,                                           /* global pointer to the active object */
                  ++prio,                                           /* priority */
                  l_mmiQueueSto, Q_DIM(l_mmiQueueSto),              /* event queue */
                  (void*)0, 0,                                      /* no per-thread stack */
                  (QEvent*)0);                                      /* no initialization event */

    QActive_start(AO_DsmApi,                                        /* global pointer to the active object */
                  ++prio,                                           /* priority */
                  l_dsmApiQueueSto, Q_DIM(l_dsmApiQueueSto),        /* event queue */
                  (void*)0, 0,                                      /* no per-thread stack */
                  (QEvent*)0);                                      /* no initialization event */

    QActive_start(AO_Datamodel,                                     /* global pointer to the active object */
                  ++prio,                                           /* priority */
                  l_datamodelQueueSto, Q_DIM(l_datamodelQueueSto),  /* event queue */
                  (void*)0, 0,                                      /* no per-thread stack */
                  (QEvent*)0);                                      /* no initialization event */

    QActive_start(AO_SelectB,                                       /* global pointer to the active object */
                  ++prio,                                           /* priority */
                  l_selectBQueueSto, Q_DIM(l_selectBQueueSto),      /* event queue */
                  (void*)0, 0,                                      /* no per-thread stack */
                  (QEvent*)0);                                      /* no initialization event */

    QActive_start(AO_SelectA,                                       /* global pointer to the active object */
                  ++prio,                                           /* priority */
                  l_selectAQueueSto, Q_DIM(l_selectAQueueSto),      /* event queue */
                  (void*)0, 0,                                      /* no per-thread stack */
                  (QEvent*)0);                                      /* no initialization event */

    QActive_start(AO_DS485,                                         /* global pointer to the active object */
                  ++prio,                                           /* priority */
                  l_ds485QueueSto, Q_DIM(l_ds485QueueSto),          /* event queue */
                  (void*)0, 0,                                      /* no per-thread stack */
                  (QEvent*)0);                                      /* no initialization event */

    // convert port // listenPortDS485,
    int port = -1;
    char *err = NULL;
    errno = 0;
    port = strtol(listenPortDS485, &err, 10);
    if ((port == 0) && (*err)) {
        vdsm_report(LOG_ERR, "Invalid listen port specified\n");
        exit(EXIT_FAILURE);
    }
    if ((port < 0) || (port > USHRT_MAX)) {
        vdsm_report(LOG_ERR, "Listen port value out of range\n");
        exit(EXIT_FAILURE);
    }

#ifdef HAVE_AVAHI
    vdc_whitelist_hash_t *allowed_devices = NULL;
    if (cfg) {
        prepare_whitelist(cfg, &allowed_devices);
    }

    /* initialize vDC connection */
    vdc_api_init(&vDSMDsuid, role, (uint16_t)port, allowed_devices);
#else
    /* initialize vDC connection */
    vdc_api_init(&vDSMDsuid, role, (uint16_t)port, NULL);
#endif

    /* run the QF application */
    QF_run();

    return 0;
}
